# Copyright (C) 2022 Hottinger Bruel and Kjaer Benelux B.V.
# Schutweg 15a
# 5145 NP Waalwijk
# The Netherlands
# http://www.hbm.com

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""GHS API States"""

RETURN_KEY = "GHSReturnValue"

# All possible API return values
GHSReturnValue = {
    "Reserved": 0,
    "OK": 1,
    "NOK": 2,
    "EmptySlot": 3,
    "NullPtrArgument": 4,
    "InvalidSlotID": 5,
    "SystemNotIdle": 6,
    "SystemNotRecording": 7,
    "SystemNotPaused": 8,
    "InvalidSampleRate": 9,
    "InvalidHandle": 10,
    "APIMismatch": 11,
    "ConnectionFailed": 12,
    "InvalidIP": 13,
    "MainframeTimeout": 14,
    "InsufficientDiskSpace": 15,
    "CreateRecordingFailed": 16,
    "NoConnection": 17,
    "IncompatibleStorage": 18,
    "RecordingNotFound": 19,
    "SystemNotInPreview": 20,
    "AlreadyConnected": 21,
    "InvalidRecordingName": 22,
    "InvalidChannelIndex": 23,
    "InvalidUserMode": 24,
    "InvalidChannelType": 25,
    "InvalidTriggerPosition": 26,
    "InvalidSweepMode": 27,
    "NoRecordersInMainframe": 28,
    "InvalidContinuousMode": 29,
    "InvalidModeForTriggerPosition": 30,
    "Adapted": 31,
    "InvalidUTF8Character": 32,
    "DuplicateChannelName": 33,
    "InvalidDataType": 34,
    "MethodNotFound": 35,
    "InvalidJSONFormat": 36,
    "UnkownErrorMessage": 37,
    "FieldBusError_FieldBusAlready_NotConfigured": 38,
    "FieldBusError_FieldBusAlready_Enabled": 39,
    "FieldBusError_NoFormulasDeployed" : 43,
    "FieldBusError_BufferSizeInvalid" : 44,
    "CANBusNotFound": 46,
    "WriteAccessBlocked": 47,
    "InvalidOutputNumber": 49,
    "IncompatibleDigitalOutputMode": 50,
}

GHSAccess = {
    "ReadOnly": 0,
    "ReadWrite": 1,
}

GHSAcquisitionState = {
    "Reserved": 0,
    "Recording": 1,
    "Pause": 2,
    "SavingData": 3,
    "Idle": 4,
    "Preview": 5,
}

GHSSyncStatus = {
    "Reserved": 0,
    "NotSynced": 1,
    "Syncing": 2,
    "Synced": 3,
    "ReSyncing": 4,
    "NoSignal": 5,
    "CoarseSynced": 6,
    "NoGMR1000": 7,
    "NoOTMC100": 8,
}

GHSUserMode = {
    "Reserved": 0,
    "Sweeps": 1,
    "Continuous": 2,
    "Dual": 3,
}

GHSStorageLocation = {
    "Reserved": 0,
    "Remote": 1,
    "Local1": 2,
    "Local2": 3,
    "iSCSI1": 4,
    "iSCSI2": 5,
}

GHSSweepRecordingMode = {"Reserved": 0, "Normal": 1, "PreTrigger": 2}

GHSContinuousRecordingMode = {
    "Reserved": 0,
    "Standard": 1,
    "Circular": 2,
    "Limited": 3,
    "StopOnTrigger": 4,
}

GHSChannelType = {
    "Invalid": 0,
    "Analog": 1,
    "Event": 2,
    "TimerCounter": 3,
}

GHSAmplifierMode = {
    "None": -1,
    "Basic": 0,
    "Bridge": 1,
    "Icp": 2,
    "ThermoCouple": 3,
    "BasicSensor": 4,
    "Charge": 5,
    "Current4_20": 6,
    "ThermoResistor": 7,
}

GHSExcitationType = {
    "Voltage": 0,
    "Voltage_Sense": 1,
    "Current": 2,
    "Voltage_Strobed": 3,
    "Voltage_Sense_Strobed": 4,
    "Current_Strobed": 5,
}

GHSFilterType = {
    "Bessel": 0,
    "Butterworth": 1,
    "Elliptic": 2,
    "FIR": 3,
    "IIR": 4,
    "Wideband": 5,
    "Bessel_AA": 6,
    "Butterworth_AA": 7,
    "SigmaDeltaWB": 8,
    "SigmaDelta": 9,
    "BandPass": 10,
    "FIR3dB": 11,
}

GHSInputCoupling = {
    "SingleEndedPositive": 0,
    "SingleEndedNegative": 1,
    "Differential": 2,
    "Current": 3,
    "FloatingDifferential": 4,
}

GHSSignalCoupling = {
    "GND": 0,
    "DC": 1,
    "AC": 2,
    "DC_RMS": 3,
    "AC_RMS": 4,
    "DC_Frequency": 5,
    "AC_Frequency": 6,
    "DC_TrueRMS": 7,
    "AC_TrueRMS": 8,
    "DC_ExternalProbe": 9,
    "AC_ExternalProbe": 10,
    "Reference": 11,
    "ZeroSet": 12,
    "SinglePrecision": 13,
    "DoublePrecision": 14,
    "QuadPrecision": 15,
    "Charge": 16,
}

GHSTriggerMode = {
    "Off": 0,
    "Basic": 1,
    "Dual": 2,
    "Window": 3,
    "DualWindow": 4,
    "Sequential": 5,
    "QualifierBasic": 6,
    "QualifierDual": 7,
}

GHSDirection = {"RisingEdge": 0, "FallingEdge": 1}

GHSTimerCounterMode = {
    "RPMUniDirectional": 0,
    "RPMBiDirectional": 1,
    "RPMQuadrature": 2,
    "FrequencyUniDirectional": 3,
    "FrequencyBiDirectional": 4,
    "FrequencyQuadrature": 5,
    "CountUniDirectional": 6,
    "CountBiDirectional": 7,
    "CountQuadrature": 8,
    "AngleQuadrature": 9,
    "AngleQuadratureWithRefPos": 10,
    "AngleUniDirectional": 11,
    "AngleUniDirectionalWithRefPos": 12,
    "AngleBiDirectional": 13,
    "AngleBiDirectionalWithRefPos": 14,
}

GHSEnableDisable = {
    "Disable": 0,
    "Enable": 1,
}

GHSRecordingDataSource = {
    "SyncChannels": 0,
    "SyncRealTimeFormulas": 1,
}

GHSDigitalOutput = {
    "Output1": 0,
    "Output2": 1,
}

GHSDigitalOutMode = {
    "Low": 0,
    "High": 1,
    "Acquiring": 2,
    "Trigger": 3,
    "Alarm": 4,
}


def to_string(value: int, ghs_dict: dict) -> str:
    """Get status key by value from dictionary."""

    for string_val, return_val in ghs_dict.items():
        if value == return_val:
            return string_val
    if ghs_dict == "GHSChannelType":
        return "Invalid"
    return "Reserved"


def from_string(key: str, ghs_dict: dict) -> int:
    """Get status value by key from dictionary."""

    for string_val, return_val in ghs_dict.items():
        if key == string_val:
            return return_val
    return 0


# TODO: Formating down from here
def getDictEntryFromValue(value, ghsDict):
    for stringReturnVal, returnVal in ghsDict.items():
        if value == returnVal:
            return {stringReturnVal: value}


def getDictEntryFromKey(key, ghsDict):
    for stringReturnVal, returnVal in ghsDict.items():
        if key == stringReturnVal:
            return {key: returnVal}



# Copyright (C) 2022 Hottinger Bruel and Kjaer Benelux B.V.
# Schutweg 15a
# 5145 NP Waalwijk
# The Netherlands
# http://www.hbm.com

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Manage recordings module interface."""

from .connection import ConnectionHandler
from .ghsapi_states import (
    RETURN_KEY,
    GHSEnableDisable,
    GHSRecordingDataSource,
    GHSReturnValue,
    GHSStorageLocation,
    from_string,
    to_string,
)


def delete_all_recordings(con_handle: ConnectionHandler) -> str:
    """Interface to deletes all recordings from local mainframe storage

    Recordings are deleted asynchronously.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
       String value representing request status.
    """

    response_json = con_handle.send_request_wait_response(
        "DeleteAllRecordings", None
    )
    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def delete_last_recording(con_handle: ConnectionHandler) -> str:
    """Interface to delete the most recent recording from local
    mainframe storage.

    Recordings are deleted asynchronously.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
       String value representing request status.
    """

    response_json = con_handle.send_request_wait_response(
        "DeleteLastRecording", None
    )
    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_recording_name(
    con_handle: ConnectionHandler,
) -> tuple[str, str | None, int | None]:
    """Interface to get recording base name and recording index of the
    last recording file.

    The recording base name parameter is UTF-8 encoded.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        Tuple with status, base name and index of the recording file.
    """

    response_json = con_handle.send_request_wait_response(
        "GetRecordingName", None
    )

    if (
        not any(
            key in response_json
            for key in [
                "RecordingIndex",
                "RecordingName",
            ]
        )
    ) or (response_json[RETURN_KEY] != GHSReturnValue["OK"]):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
            None,
        )
    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        response_json["RecordingName"],
        response_json["RecordingIndex"],
    )


def set_recording_name(
    con_handle: ConnectionHandler, recording_name: str, recording_index: int
) -> str:
    """Interface to set the recording base name and recording index for
    the next recording file.

    The system needs to be idle before calling this function.
    The recording base name parameter must be UTF-8 encoded.

    Args:
        con_handle: A unique identifier per mainframe connection.
        recording_name: The desired base name of the recording file.
        recording_index: The desired index of the recording file.

    Returns:
       String value representing request status.
    """

    if not recording_name or not recording_index:
        return "NullPtrArgument"

    recording_name_dict = {
        "RecordingBaseName": recording_name,
        "RecordingIndex": recording_index,
    }

    response_json = con_handle.send_request_wait_response(
        "SetRecordingName", recording_name_dict
    )
    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_storage_location(
    con_handle: ConnectionHandler,
) -> tuple[str, str | None]:
    """Interface to retrieve the storage location.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
       Tuple with status and storage location.
    """

    response_json = con_handle.send_request_wait_response(
        "GetStorageLocation", None
    )

    if ("StorageLocation" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return to_string(response_json[RETURN_KEY], GHSReturnValue), None
    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["StorageLocation"], GHSStorageLocation),
    )


def set_storage_location(
    con_handle: ConnectionHandler, storage_location: str | int
) -> str:
    """Interface to set the storage location.

    The system needs to be idle before calling this function.

    Args:
        con_handle: A unique identifier per mainframe connection.
        storage_location: The desired storage location.

    Returns:
       String value representing request status.
    """

    if not storage_location:
        return "NullPtrArgument"

    if (
        isinstance(storage_location, str)
        and storage_location in GHSStorageLocation
    ):
        storage_loc_dict = {
            "StorageLocation": from_string(
                storage_location, GHSStorageLocation
            )
        }

    elif (
        isinstance(storage_location, int)
        and storage_location in GHSStorageLocation.values()
    ):
        storage_loc_dict = {"StorageLocation": storage_location}

    else:
        return "IncompatibleStorage"

    response_json = con_handle.send_request_wait_response(
        "SetStorageLocation", storage_loc_dict
    )
    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_high_low_rate_storage_enabled(
    con_handle: ConnectionHandler,
    source: str | int,
    slot_id: str,
) -> tuple[str, str | None, str | None]:
    """Interface to retrieve storage enabled status of high and low
    rate data for the specified recording data source.

    Args:
        con_handle: A unique identifier per mainframe connection.
        source: The recording data source to retrieve the storage
        enabled status for.
        slot_id: The slot containing the recorder to retrieve the
        storage enabled status for (e.g. 'A' for the first slot). This
        argument is only evaluated for recorder based data sources.

    Returns:
        Tuple with status, flag to indicate if high and low rate data
        is stored.
    """

    if not source or not slot_id:
        return "NullPtrArgument"

    if isinstance(source, str) and source in GHSRecordingDataSource:
        source_slot_dict = {
            "Source": from_string(source, GHSRecordingDataSource),
            "SlotId": slot_id,
        }

    elif isinstance(source, int) and source in GHSRecordingDataSource.values():
        source_slot_dict = {"Source": source, "SlotId": slot_id}

    else:
        return "InvalidDataType"

    response_json = con_handle.send_request_wait_response(
        "GetHighLowRateStorageEnabled", source_slot_dict
    )

    if (
        not any(
            key in response_json
            for key in [
                "HighRateEnable",
                "LowRateEnable",
            ]
        )
    ) or (response_json[RETURN_KEY] != GHSReturnValue["OK"]):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
            None,
        )
    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["HighRateEnable"], GHSEnableDisable),
        to_string(response_json["LowRateEnable"], GHSEnableDisable),
    )


def set_high_low_rate_storage_enabled(
    con_handle: ConnectionHandler,
    source: str | int,
    slot_id: str,
    high_rate_enabled: str | int,
    low_rate_enabled: str | int,
) -> str:
    """Interface to enable/disable storage of high and low rate data
    for the specified recording data source.

    The system needs to be idle before calling this function.

    Args:
        con_handle: A unique identifier per mainframe connection.
        source: The recording data source to retrieve the storage
        enabled status for.
        slot_id: The slot containing the recorder to retrieve the
        storage enabled status for (e.g. 'A' for the first slot). This
        argument is only evaluated for recorder based data sources.
        high_rate_enabled: Enable/disable storage of high rate data.
        low_rate_enabled: Enable/disable storage of low rate data.

    Returns:
        String value representing request status.
    """

    if (
        not source
        or not slot_id
        or not high_rate_enabled
        or not low_rate_enabled
    ):
        return "NullPtrArgument"

    if isinstance(source, str) and source in GHSRecordingDataSource:
        source = from_string(source, GHSRecordingDataSource)

    elif isinstance(source, int) and source in GHSRecordingDataSource.values():
        pass

    else:
        return "InvalidDataType"

    if (
        isinstance(high_rate_enabled, str)
        and high_rate_enabled in GHSEnableDisable
    ):
        high_rate_enabled = from_string(high_rate_enabled, GHSEnableDisable)

    elif (
        isinstance(high_rate_enabled, int)
        and high_rate_enabled in GHSEnableDisable.values()
    ):
        pass

    else:
        return "InvalidDataType"

    if (
        isinstance(low_rate_enabled, str)
        and low_rate_enabled in GHSEnableDisable
    ):
        low_rate_enabled = from_string(low_rate_enabled, GHSEnableDisable)

    elif (
        isinstance(low_rate_enabled, int)
        and low_rate_enabled in GHSEnableDisable.values()
    ):
        pass

    else:
        return "InvalidDataType"

    source_slot_dict = {
        "Source": source,
        "SlotId": slot_id,
        "HighRateEnable": high_rate_enabled,
        "LowRateEnable": low_rate_enabled,
    }

    response_json = con_handle.send_request_wait_response(
        "SetHighLowRateStorageEnabled", source_slot_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)



# Copyright (C) 2022 Hottinger Bruel and Kjaer Benelux B.V.
# Schutweg 15a
# 5145 NP Waalwijk
# The Netherlands
# http://www.hbm.com

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Connection module interface.

It is used to connect disconnect to the mainframe.
"""

from .connection import ConnectionHandler
from .ghsapi_states import RETURN_KEY, GHSAccess, GHSReturnValue, to_string


def connect(
    con_handle: ConnectionHandler,
    ip_address: int,
    port_num: int,
    client_api_version: int,
) -> str:
    """Interface to connect to the mainframe.

    Args:
        con_handle: A unique identifier per mainframe connection.
        ip_address: Mainframe ip address.
        port_num: Mainframe port number.
        client_api_version: Client supported API version.

    Returns:
        String value representing connect request status.
    """

    if (
        not con_handle
        or not ip_address
        or not port_num
        or not client_api_version
    ):
        return "NullPtrArgument"

    return_var = con_handle.connection_establish(ip_address, port_num)
    if return_var != GHSReturnValue["OK"]:
        return to_string(return_var, GHSReturnValue)

    connect_param_dict = {"ClientAPIVersion": client_api_version}
    response_json = con_handle.send_request_wait_response(
        "Connect", connect_param_dict
    )
    if response_json[RETURN_KEY] != GHSReturnValue["OK"]:
        try:
            if response_json["ServerAPIVersion"] != client_api_version:
                print(
                    f"ServerAPIVersion: {response_json['ServerAPIVersion']} \
                        ClientAPIVersion: {client_api_version}"
                )
        except KeyError:
            pass
        return to_string(response_json[RETURN_KEY], GHSReturnValue)

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_current_access(con_handle: ConnectionHandler) -> str:
    """Interface to get current access permission.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        String representing current access permissions for client.
    """

    response_json = con_handle.send_request_wait_response(
        "GetCurrentAccess", None
    )

    try:
        return to_string(response_json["Access"], GHSAccess)
    except KeyError:
        return to_string(response_json[RETURN_KEY], GHSReturnValue)


def disconnect(con_handle: ConnectionHandler) -> str:
    """Interface to disconnect to the mainframe.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        String value representing disconnect request status.
    """

    response_json = con_handle.send_request_wait_response("Disconnect", None)
    return to_string(response_json[RETURN_KEY], GHSReturnValue)



# Copyright (C) 2022 Hottinger Bruel and Kjaer Benelux B.V.
# Schutweg 15a
# 5145 NP Waalwijk
# The Netherlands
# http://www.hbm.com

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Fieldbus module interface."""

from .connection import ConnectionHandler
from .ghsapi_states import (
    RETURN_KEY,
    GHSReturnValue,
    to_string,
)

    
FLOAT_PRECISION = 3


# Functions
def get_fieldbus_data_name_and_unit(
    con_handle: ConnectionHandler, data_index: int
) -> tuple[str, str | None, str | None]:
    """Gets the name of specific field bus data value, by giving the index of the data value.
    
    *The number of data values can be retrieved by ghs_get_fieldbus_data_count().
    
    *In case of no formulas (or reserved values acquisition state and latency) published,
    an error code is returned: FieldBusError_NoFormulasDeployed.
    
    *Read - This method can be called by multiple connected clients at same
    time.*
    
    Args:
        con_handle: A unique identifier per mainframe connection.
        data_index: Zero-based index of the data value.
    
    Returns:
       Tuple with status, data name and data unit for the desired fieldbus index.
    """

    if data_index < 0 or data_index is None:
        return "NullPtrArgument", None, None

    fieldbus_type_dict = {
        "DataIndex": data_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetFieldBusDataNameAndUnit", fieldbus_type_dict
    )

    if ("DataName" not in response_json) or ("DataUnit" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return to_string(response_json[RETURN_KEY], GHSReturnValue), None, None

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        response_json["DataName"],
        response_json["DataUnit"]
    )


def get_fieldbus_data_count(
    con_handle: ConnectionHandler
) -> tuple[str, int | None]:
    """Gets the number of data that are configured to be sent through the field bus, excluding the timestamp.
    
    *The number of data values can be retrieved by ghs_get_fieldbus_data_count().
    
    *Read - This method can be called by multiple connected clients at same
    time.*

    Args:
        con_handle: A unique identifier per mainframe connection.
    
    Returns:
       Tuple with status and total number of fieldbus data channel being published.
    """

    response_json = con_handle.send_request_wait_response(
        "GetFieldBusDataCount", None
    )

    if ("DataCount" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return to_string(response_json[RETURN_KEY], GHSReturnValue), None

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        response_json["DataCount"],
    )

def get_fieldbus_request_snapshot(
    con_handle: ConnectionHandler,
    data_count: int,
) -> tuple[str, list[float] | None, int | None, list[float] | None]:
    """ Request a single snapshot of the field bus data.
    
    *In this case the timestamp received is not grid aligned.
    
    *If the system is not acquiring the method returns an error code: SystemNotRecording
    
    *In case of no formulas you get the error return code back: FieldBusError_NotConfigured
    
    *If the field bus is already enabled then this method is not allowed. 
    An error return code is returned: FieldBusError_FieldBusAlready_Enabled
    
    *Read - This method can be called by multiple connected clients at same
    time.*

    Args:
        con_handle: A unique identifier per mainframe connection.
        data_index: Zero-based index of the data value.
    
    Returns:
       Tuple with status, timestamp, total number of fieldbus channels and the fieldbus data.
    """

    if data_count < 0 or data_count is None:
        return "NullPtrArgument", None, None, None

    response_json = con_handle.send_request_wait_response(
        "RequestSnapshot", None
    )

    if ("DataCount" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return to_string(response_json[RETURN_KEY], GHSReturnValue), None, None, None

    if (data_count != response_json["DataCount"]):
        return "FieldBusError_BufferSizeInvalid", None, None, None

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        round(response_json["TimeStamp"], FLOAT_PRECISION),
        response_json["DataCount"],
        [round(channel_data, FLOAT_PRECISION) for channel_data in response_json["Data"]]
    )



# Copyright (C) 2022 Hottinger Bruel and Kjaer Benelux B.V.
# Schutweg 15a
# 5145 NP Waalwijk
# The Netherlands
# http://www.hbm.com

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Channel module interface."""

from .connection import ConnectionHandler
from .ghsapi_states import (
    RETURN_KEY,
    GHSAmplifierMode,
    GHSChannelType,
    GHSDirection,
    GHSEnableDisable,
    GHSExcitationType,
    GHSFilterType,
    GHSInputCoupling,
    GHSReturnValue,
    GHSSignalCoupling,
    GHSTimerCounterMode,
    GHSTriggerMode,
    from_string,
    to_string,
)

# Functions


def get_channel_type(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, str | None]:
    """Determine the type of a channel.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status and type of the channel.
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None

    channel_type_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetChannelType", channel_type_dict
    )

    if ("ChannelType" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return to_string(response_json[RETURN_KEY], GHSReturnValue), None

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["ChannelType"], GHSChannelType),
    )


def get_channel_name(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, str | None]:
    """Determine the name of a channel.

     The channel name is UTF-8 encoded.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status and name of the channel.
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None

    channel_name_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetChannelName", channel_name_dict
    )

    if ("ChannelName" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return to_string(response_json[RETURN_KEY], GHSReturnValue), None

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        response_json["ChannelName"],
    )


def set_channel_name(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    channel_name: str,
) -> str:
    """Set the name of a channel.

     The channel name is UTF-8 encoded.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        channel_name: The desired channel name.

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not channel_name:
        return "NullPtrArgument"

    channel_name_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "ChannelName": channel_name,
    }

    response_json = con_handle.send_request_wait_response(
        "SetChannelName", channel_name_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_channel_storage_enabled(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, str | None]:
    """Determine if storage is enabled or disabled for a channel.

    Read - This method can be called by multiple connected clients at
    same time

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status and storage enabled status for the channel.
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None

    channel_enabled_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetChannelStorageEnabled", channel_enabled_dict
    )

    if ("Enabled" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return to_string(response_json[RETURN_KEY], GHSReturnValue), None

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["Enabled"], GHSEnableDisable),
    )


def set_channel_storage_enabled(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    enabled: str | int,
) -> str:
    """Enable or disable storage for a channel.

    The system needs to be idle before calling this function.

    ReadWrite - This method will only process requests from the
    connected client with the most privileges order (Privileges order:
    1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        enabled: The desired storage enabled status for the channel

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not enabled:
        return "NullPtrArgument"

    if isinstance(enabled, str) and enabled in GHSEnableDisable:
        enabled = from_string(enabled, GHSEnableDisable)

    elif isinstance(enabled, int) and enabled in GHSEnableDisable.values():
        pass

    else:
        return "InvalidDataType"

    channel_enabled_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "ChannelStorageEnable": enabled,
    }

    response_json = con_handle.send_request_wait_response(
        "SetChannelStorageEnabled", channel_enabled_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def cmd_zeroing(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    ezeroing: str | int,
) -> str:
    """Perform zeroing in a channel.

     The system needs to be idle before calling this function.

     Read - This method can be called by multiple connected clients at
     same time

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        ezeroing: Zero / Unzero the specific channel.

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not ezeroing:
        return "NullPtrArgument"

    if isinstance(ezeroing, str) and ezeroing in GHSEnableDisable:
        ezeroing = from_string(ezeroing, GHSEnableDisable)

    elif isinstance(ezeroing, int) and ezeroing in GHSEnableDisable.values():
        pass

    else:
        return "InvalidDataType"

    ezeroing_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "ZeroingMode": ezeroing,
    }

    response_json = con_handle.send_request_wait_response(
        "Zeroing", ezeroing_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


## Modules

# Analog


def get_trigger_settings(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[
    str, str | None, float | None, float | None, float | None, str | None
]:
    """Determine the trigger settings for an analog channel.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status and trigger settings.
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None, None, None, None, None

    trigger_settings_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetTriggerSettings", trigger_settings_dict
    )

    if (
        not any(
            key in response_json
            for key in [
                "TriggerMode",
                "PrimaryLevel",
                "SecondaryLevel",
                "Hysteresis",
                "Direction",
            ]
        )
    ) or (response_json[RETURN_KEY] != GHSReturnValue["OK"]):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
            None,
            None,
            None,
            None,
        )

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["TriggerMode"], GHSTriggerMode),
        response_json["PrimaryLevel"],
        response_json["SecondaryLevel"],
        response_json["Hysteresis"],
        to_string(response_json["Direction"], GHSDirection),
    )


def set_trigger_settings(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    trigger_mode: str | int,
    primary_level: float,
    secondary_level: float,
    hysteresis: float,
    direction: str | int,
) -> str:
    """Set the trigger settings for an analog channel.

     The system needs to be idle before calling this function.

     This function overwrites any previously set trigger settings for
     the specified recorder.

     If the specified trigger mode or value is not supported by the
     recorder, the trigger mode remains unchanged or the value is
     rounded to the nearest supported value.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        trigger_mode: Trigger Mode. Default is TriggerMode_Basic
        primary_level: The desired primary trigger level.
        secondary_level: The desired secondary trigger level.
        hysteresis: The desired trigger hysteresis.
        direction: The desired trigger direction.

    Returns:
       String value representing request status.
    """

    if (
        not slot_id
        or not channel_index
        or not trigger_mode
        or not primary_level
        or not secondary_level
        or not hysteresis
        or not direction
    ):
        return "NullPtrArgument"

    if isinstance(trigger_mode, str) and trigger_mode in GHSTriggerMode:
        trigger_mode = from_string(trigger_mode, GHSTriggerMode)

    elif (
        isinstance(trigger_mode, int)
        and trigger_mode in GHSTriggerMode.values()
    ):
        pass

    else:
        return "InvalidDataType"

    if isinstance(direction, str) and direction in GHSDirection:
        direction = from_string(direction, GHSDirection)

    elif isinstance(direction, int) and direction in GHSDirection.values():
        pass

    else:
        return "InvalidDataType"

    trigger_settings_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "TriggerMode": trigger_mode,
        "PrimaryLevel": primary_level,
        "SecondaryLevel": secondary_level,
        "Hysteresis": hysteresis,
        "Direction": direction,
    }

    response_json = con_handle.send_request_wait_response(
        "SetTriggerSettings", trigger_settings_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_signal_coupling(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, str | None]:
    """Determine the signal coupling for an analog channel.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status and signal coupling for an analog channel.
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None

    signal_coupling_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetSignalCoupling", signal_coupling_dict
    )

    if ("SignalCoupling" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return to_string(response_json[RETURN_KEY], GHSReturnValue), None

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["SignalCoupling"], GHSSignalCoupling),
    )


def set_signal_coupling(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    signal_coupling: str | int,
) -> str:
    """Set the signal coupling for an analog channel.

     If the specified signal coupling mode is not supported by the
     recorder, the signal coupling mode remains unchanged.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        signal_coupling: The desired signal coupling.

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not signal_coupling:
        return "NullPtrArgument"

    if (
        isinstance(signal_coupling, str)
        and signal_coupling in GHSSignalCoupling
    ):
        signal_coupling = from_string(signal_coupling, GHSSignalCoupling)

    elif (
        isinstance(signal_coupling, int)
        and signal_coupling in GHSSignalCoupling.values()
    ):
        pass

    else:
        return "InvalidDataType"

    signal_coupling_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "SignalCoupling": signal_coupling,
    }

    response_json = con_handle.send_request_wait_response(
        "SetSignalCoupling", signal_coupling_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_input_coupling(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, str | None]:
    """Determine the input coupling for an analog channel.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status and input coupling for an analog channel.
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None

    input_coupling_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetInputCoupling", input_coupling_dict
    )

    if ("InputCoupling" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return to_string(response_json[RETURN_KEY], GHSReturnValue), None

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["InputCoupling"], GHSInputCoupling),
    )


def set_input_coupling(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    input_coupling: str | int,
) -> str:
    """Set the input coupling for an analog channel.

     The system needs to be idle before calling this function.

     If the specified input coupling mode is not supported by the
     recorder, the input coupling mode remains unchanged.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        input_coupling: The desired input coupling.

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not input_coupling:
        return "NullPtrArgument"

    if isinstance(input_coupling, str) and input_coupling in GHSInputCoupling:
        input_coupling = from_string(input_coupling, GHSInputCoupling)

    elif (
        isinstance(input_coupling, int)
        and input_coupling in GHSInputCoupling.values()
    ):
        pass

    else:
        return "InvalidDataType"

    input_coupling_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "InputCoupling": input_coupling,
    }

    response_json = con_handle.send_request_wait_response(
        "SetInputCoupling", input_coupling_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_span_and_offset(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, float | None, float | None]:
    """Determine the span and offset for an analog channel.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status, span and offset for an analog channel.
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None, None

    span_offset_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetSpanAndOffset", span_offset_dict
    )

    if (
        not any(
            key in response_json
            for key in [
                "Span",
                "Offset",
            ]
        )
    ) or (response_json[RETURN_KEY] != GHSReturnValue["OK"]):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
            None,
        )

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        response_json["Span"],
        response_json["Offset"],
    )


def set_span_and_offset(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    span: float,
    offset: float,
) -> str:
    """Set Span and offset for analog channels.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        span: The span in user units. The value is adapted to available
        options.
        offset: The offset in user units. The value is adapted to
        available options.

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not span or not offset:
        return "NullPtrArgument"

    if not (isinstance(span, float) and isinstance(offset, float)):
        return "InvalidDataType"

    span_offset_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "Span": span,
        "Offset": offset,
    }

    response_json = con_handle.send_request_wait_response(
        "SetSpanAndOffset", span_offset_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_filter_type_and_frequency(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, str | None, float | None]:
    """Determine the filter type and frequency for an analog channel.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status, filter type and frequency for an analog
       channel.
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None, None

    filter_freq_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetFilterTypeAndFrequency", filter_freq_dict
    )

    if (
        not any(
            key in response_json
            for key in [
                "FilterType",
                "Frequency",
            ]
        )
    ) or (response_json[RETURN_KEY] != GHSReturnValue["OK"]):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
            None,
        )

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["FilterType"], GHSFilterType),
        response_json["Frequency"],
    )


def set_filter_type_and_frequency(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    filter_type: str | int,
    frequency: float,
) -> str:
    """Set the filter type and frequency for an analog channel.

     This function overwrites any previously set filter settings for
     the specified recorder.

     If the specified filter type or value is not supported by the
     recorder, the filter type remains unchanged or the value is
     rounded to the nearest supported value.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        filter_type: The filter type. Default is GHSFilterType_Bessel.
        frequency: The filter frequency in Hz.

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not filter_type or not frequency:
        return "NullPtrArgument"

    if isinstance(filter_type, str) and filter_type in GHSFilterType:
        filter_type = from_string(filter_type, GHSFilterType)

    elif (
        isinstance(filter_type, int) and filter_type in GHSFilterType.values()
    ):
        pass

    else:
        return "InvalidDataType"

    if not isinstance(frequency, float):
        return "InvalidDataType"

    filter_freq_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "FilterType": filter_type,
        "Frequency": frequency,
    }

    response_json = con_handle.send_request_wait_response(
        "SetFilterTypeAndFrequency", filter_freq_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_excitation(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, str | None, float | None]:
    """Determine the excitation type and value for an analog channel.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status, excitation type and excitation value
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None, None

    excitation_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetExcitation", excitation_dict
    )

    if (
        not any(
            key in response_json
            for key in [
                "ExcitationType",
                "ExcitationValue",
            ]
        )
    ) or (response_json[RETURN_KEY] != GHSReturnValue["OK"]):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
            None,
        )

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["ExcitationType"], GHSExcitationType),
        response_json["ExcitationValue"],
    )


def set_excitation(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    excitation_type: str | int,
    excitation_value: float,
) -> str:
    """Set the excitation type and value for an analog channel.

     The system needs to be idle before calling this function.

     If the specified excitation type or value is not supported by the
     recorder, the excitation type remains unchanged or the value is rounded
     to the nearest supported value.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        excitation_type: The desired excitation type. Default is
        ExcitationType_Voltage.
        excitation_value: The desired excitation value in user units (voltage
        or current).

    Returns:
       String value representing request status.
    """

    if (
        not slot_id
        or not channel_index
        or not excitation_type
        or not excitation_value
    ):
        return "NullPtrArgument"

    if (
        isinstance(excitation_type, str)
        and excitation_type in GHSExcitationType
    ):
        excitation_type = from_string(excitation_type, GHSExcitationType)

    elif (
        isinstance(excitation_type, int)
        and excitation_type in GHSExcitationType.values()
    ):
        pass

    else:
        return "InvalidDataType"

    if not isinstance(excitation_value, float):
        return "InvalidDataType"

    excitation_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "ExcitationType": excitation_type,
        "ExcitationValue": excitation_value,
    }

    response_json = con_handle.send_request_wait_response(
        "SetExcitation", excitation_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_amplifier_mode(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, str | None]:
    """Determine the amplifier mode for an analog channel.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status and amplifier mode
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None

    amplifier_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetAmplifierMode", amplifier_dict
    )

    if ("AmplifierMode" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
        )

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["AmplifierMode"], GHSAmplifierMode),
    )


def set_amplifier_mode(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    amplifier_mode: str | int,
) -> str:
    """Set the amplifier mode for an analog channel.

     The system needs to be idle before calling this function.

     If the specified amplifier mode is not supported by the recorder, the
     amplifier mode remains unchanged.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        amplifier_mode: The desired amplifier mode

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not amplifier_mode:
        return "NullPtrArgument"

    if isinstance(amplifier_mode, str) and amplifier_mode in GHSAmplifierMode:
        amplifier_mode = from_string(amplifier_mode, GHSAmplifierMode)

    elif (
        isinstance(amplifier_mode, int)
        and amplifier_mode in GHSAmplifierMode.values()
    ):
        pass

    else:
        return "InvalidDataType"

    amplifier_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "AmplifierMode": amplifier_mode,
    }

    response_json = con_handle.send_request_wait_response(
        "SetAmplifierMode", amplifier_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_technical_units(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, str | None, float | None, float | None]:
    """Determine the technical units, unit multiplier and unit offset for an
    analog channel.

    The units parameter is UTF-8 encoded.

    Read - This method can be called by multiple connected clients at
    same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status, technical units, technical units multiplier value
       and technical units offset value.
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None, None, None

    technical_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetTechnicalUnits", technical_dict
    )

    if (
        not any(
            key in response_json
            for key in [
                "Multiplier",
                "Offset",
                "UnitType",
            ]
        )
    ) or (response_json[RETURN_KEY] != GHSReturnValue["OK"]):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
            None,
            None,
        )

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        response_json["UnitType"],
        response_json["Multiplier"],
        response_json["Offset"],
    )


def set_technical_units(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    units: str,
    multiplier: float,
    offset: float,
) -> str:
    """Set the technical units, unit multiplier and unit offset for an analog
    channel.

     The units parameter is UTF-8 encoded.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        units: The desired technical units (e.g. 'V' for Volt or 'Hz' for
        Hertz).
        multiplier: The desired technical units multiplier value.
        offset: The desired technical units offset value.

    Returns:
       String value representing request status.
    """

    if (
        not slot_id
        or not channel_index
        or not units
        or not multiplier
        or not offset
    ):
        return "NullPtrArgument"

    if (
        not isinstance(units, str)
        or not isinstance(multiplier, float)
        or not isinstance(offset, float)
    ):
        return "InvalidDataType"

    technical_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "UnitType": units,
        "Multiplier": multiplier,
        "Offset": offset,
    }

    response_json = con_handle.send_request_wait_response(
        "SetTechnicalUnits", technical_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_auto_range(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, str | None, float | None]:
    """Determine the auto range enable and time settings.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status, auto range enabled setting and time for auto range
       in seconds
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None, None

    auto_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetAutoRange", auto_dict
    )

    if (
        not any(
            key in response_json
            for key in [
                "AutoRangeEnabled",
                "AutoRangeTime",
            ]
        )
    ) or (response_json[RETURN_KEY] != GHSReturnValue["OK"]):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
            None,
        )

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["AutoRangeEnabled"], GHSEnableDisable),
        response_json["AutoRangeTime"],
    )


def set_auto_range(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    auto_range_enabled: str | int,
    auto_range_time: float,
) -> str:
    """Set Auto range settings for analog channels.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        auto_range_enabled: The auto range enabled setting. The value is
        adapted to available options.
        auto_range_time: The time for auto range in seconds. The value is
        adapted to available options.

    Returns:
       String value representing request status.
    """

    if (
        not slot_id
        or not channel_index
        or not auto_range_enabled
        or not auto_range_time
    ):
        return "NullPtrArgument"

    if (
        isinstance(auto_range_enabled, str)
        and auto_range_enabled in GHSEnableDisable
    ):
        auto_range_enabled = from_string(auto_range_enabled, GHSEnableDisable)

    elif (
        isinstance(auto_range_enabled, int)
        and auto_range_enabled in GHSEnableDisable.values()
    ):
        pass

    else:
        return "InvalidDataType"

    if not isinstance(auto_range_time, float):
        return "InvalidDataType"

    auto_range_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "AutoRangeEnabled": auto_range_enabled,
        "AutoRangeTime": auto_range_time,
    }

    response_json = con_handle.send_request_wait_response(
        "SetAutoRange", auto_range_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def cmd_auto_range_now(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    auto_range_time: float,
) -> str:
    """Command a single shot for auto range.

     The system needs to be acquiring for this function to have any effect.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        auto_range_time: The time for auto range in seconds. The value is
        adapted to available options.

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not auto_range_time:
        return "NullPtrArgument"

    if not isinstance(auto_range_time, float):
        return "InvalidDataType"

    auto_range_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "AutoRangeTime": auto_range_time,
    }

    response_json = con_handle.send_request_wait_response(
        "AutoRangeNow", auto_range_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_channel_cal_info(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[
    str, str | None, str | None, str | None, str | None, str | None, str | None
]:
    """Retrieve calibration information for an analog channel.

    The calibrationDateTime, verificationDateTime, powerVerificationDateTime,
    calibrationLab, verificationLab and powerVerificationLab parameters are
    UTF-8 encoded.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status and calibration information
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None, None, None, None, None, None

    cal_info_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetChannelCalibrationInformation", cal_info_dict
    )

    if (
        not any(
            key in response_json
            for key in [
                "CalibrationDateTime",
                "VerificationDateTime",
                "PowerVerificationDateTime",
                "CalibrationLab",
                "VerificationLab",
                "PowerVerificationLab",
            ]
        )
    ) or (response_json[RETURN_KEY] != GHSReturnValue["OK"]):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
            None,
            None,
            None,
            None,
            None,
        )

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        response_json["CalibrationDateTime"],
        response_json["VerificationDateTime"],
        response_json["PowerVerificationDateTime"],
        response_json["CalibrationLab"],
        response_json["VerificationLab"],
        response_json["PowerVerificationLab"],
    )


# Timer/Counter


def get_timer_counter_gate_time(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, float | None]:
    """Determine the gate time for a timer/counter channel.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status and the gate time in seconds.
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None

    gate_time_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetTimerCounterGateTime", gate_time_dict
    )

    if ("GateTime" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
        )

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        response_json["GateTime"],
    )


def set_timer_counter_gate_time(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    gate_time: float,
) -> str:
    """Set the gate time for a timer/counter channel.

     The system needs to be idle before calling this function.

     If the specified timer/counter gate time is not supported by the
     recorder, the timer/counter gate time is rounded to the nearest supported
     gate time.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        gate_time: The desired gate time in seconds.

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not gate_time:
        return "NullPtrArgument"

    if not isinstance(gate_time, float):
        return "InvalidDataType"

    gate_time_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "GateTime": gate_time,
    }

    response_json = con_handle.send_request_wait_response(
        "SetTimerCounterGateTime", gate_time_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_timer_counter_mode(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, str | None]:
    """Determine the mode for a timer/counter channel.

     Read - This method can be called by multiple connected clients at
     same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status and the timer/counter mode
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None

    mode_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetTimerCounterMode", mode_dict
    )

    if ("TimerCounterMode" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
        )

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        to_string(response_json["TimerCounterMode"], GHSTimerCounterMode),
    )


def set_timer_counter_mode(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    mode: str | int,
) -> str:
    """Set the mode for a timer/counter channel.

     The system needs to be idle before calling this function.

     If the specified timer/counter mode is not supported by the recorder, the
     timer/counter mode remains unchanged.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        mode: The desired timer/counter mode. Default is
        TimerCounterMode_RPMUniDirectional.

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not mode:
        return "NullPtrArgument"

    if isinstance(mode, str) and mode in GHSTimerCounterMode:
        mode = from_string(mode, GHSTimerCounterMode)

    elif isinstance(mode, int) and mode in GHSTimerCounterMode.values():
        pass

    else:
        return "InvalidDataType"

    mode_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "TimerCounterMode": mode,
    }

    response_json = con_handle.send_request_wait_response(
        "SetTimerCounterMode", mode_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_timer_counter_range(
    con_handle: ConnectionHandler, slot_id: str, channel_index: int
) -> tuple[str, float | None, float | None]:
    """Determine the range for a timer/counter channel.

    Read - This method can be called by multiple connected clients at
    same time.

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.

    Returns:
       Tuple with status, the lower range value and the upper range value.
    """

    if not slot_id or not channel_index:
        return "NullPtrArgument", None, None

    range_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
    }

    response_json = con_handle.send_request_wait_response(
        "GetTimerCounterRange", range_dict
    )

    if (
        not any(
            key in response_json
            for key in [
                "UpperValue",
                "LowerValue",
            ]
        )
    ) or (response_json[RETURN_KEY] != GHSReturnValue["OK"]):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
            None,
        )

    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        response_json["LowerValue"],
        response_json["UpperValue"],
    )


def set_timer_counter_range(
    con_handle: ConnectionHandler,
    slot_id: str,
    channel_index: int,
    lower_value: float,
    upper_value: float,
) -> str:
    """Set the range for a timer/counter channel.

     The system needs to be idle before calling this function.

     If the specified timer/counter range is illegal (i.e. upperValue <
     lowerValue), the timer/counter range values are corrected to the nearest
     possible values.

     ReadWrite - This method will only process requests from the
     connected client with the most privileges order (Privileges order:
     1- Perception, 2- GenDaq, 3- Other)

    Args:
        con_handle: A unique identifier per mainframe connection.
        slot_id: The slot containing the recorder to get number of
        channels for (e.g. 'A' for the first slot).
        channel_index: The zero-based index of the channel to determine
        the type for.
        lower_value: The desired lower range value.
        upper_value: The desired upper range value.

    Returns:
       String value representing request status.
    """

    if not slot_id or not channel_index or not lower_value or not upper_value:
        return "NullPtrArgument"

    if not isinstance(lower_value, float) or not isinstance(
        upper_value, float
    ):
        return "InvalidDataType"

    range_dict = {
        "SlotId": slot_id,
        "ChannelIndex": channel_index,
        "UpperValue": upper_value,
        "LowerValue": lower_value,
    }

    response_json = con_handle.send_request_wait_response(
        "SetTimerCounterRange", range_dict
    )

    return to_string(response_json[RETURN_KEY], GHSReturnValue)



# Copyright (C) 2022 Hottinger Bruel and Kjaer Benelux B.V.
# Schutweg 15a
# 5145 NP Waalwijk
# The Netherlands
# http://www.hbm.com

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Acquisition control module interface.

It is used to control acquisition state of the mainframe.
"""

from .connection import ConnectionHandler
from .ghsapi_states import (
    RETURN_KEY,
    GHSAcquisitionState,
    GHSReturnValue,
    to_string,
)


def start_preview(con_handle: ConnectionHandler) -> str:
    """Interface to start preview mode.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        String value representing request status.
    """

    response_json = con_handle.send_request_wait_response("StartPreview", None)
    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def stop_preview(con_handle: ConnectionHandler) -> str:
    """Interface to stop preview mode.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        An String value representing request status.
    """

    response_json = con_handle.send_request_wait_response("StopPreview", None)
    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def start_recording(con_handle: ConnectionHandler) -> str:
    """Interface to start recording on local storage.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        String value representing request status.
    """

    response_json = con_handle.send_request_wait_response(
        "StartRecording", None
    )
    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def pause_recording(con_handle: ConnectionHandler) -> str:
    """Interface to pause a started recording.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        String value representing request status.
    """

    response_json = con_handle.send_request_wait_response(
        "PauseRecording", None
    )
    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def resume_recording(con_handle: ConnectionHandler) -> str:
    """Interface to resume a paused recording.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        String value representing request status.
    """

    response_json = con_handle.send_request_wait_response(
        "ResumeRecording", None
    )
    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def stop_recording(con_handle: ConnectionHandler) -> str:
    """Interface to stop a started recording.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        String value representing request status.
    """

    response_json = con_handle.send_request_wait_response(
        "StopRecording", None
    )
    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def trigger(con_handle: ConnectionHandler) -> str:
    """Interface to issue a trigger.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        String value representing request status.
    """

    response_json = con_handle.send_request_wait_response("Trigger", None)
    return to_string(response_json[RETURN_KEY], GHSReturnValue)


def get_acquisition_state(
    con_handle: ConnectionHandler,
) -> tuple[str, str | None]:
    """Interface to get acquisition state of mainframe.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        Tuple with status and acquisition state of the mainframe.
    """

    response_json = con_handle.send_request_wait_response(
        "GetAcquisitionState", None
    )
    if ("GHSAcquisitionState" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return to_string(response_json[RETURN_KEY], GHSReturnValue), None
    return to_string(response_json[RETURN_KEY], GHSReturnValue), to_string(
        response_json["GHSAcquisitionState"], GHSAcquisitionState
    )


def get_acquisition_start_time(
    con_handle: ConnectionHandler,
) -> tuple[str, int | None, int | None, float | None]:
    """Interface to get absolute time of the start of acquisition.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
        Tuple with status and the absolute time of the start of
        acquisition.
    """

    response_json = con_handle.send_request_wait_response(
        "GetAcquisitionStartTime", None
    )

    if (
        not any(
            key in response_json
            for key in [
                "AbsoluteTimeYear",
                "AbsoluteTimeDay",
                "AbsoluteTimeSeconds",
            ]
        )
    ) or (response_json[RETURN_KEY] != GHSReturnValue["OK"]):
        return (
            to_string(response_json[RETURN_KEY], GHSReturnValue),
            None,
            None,
            None,
        )
    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        response_json["AbsoluteTimeYear"],
        response_json["AbsoluteTimeDay"],
        response_json["AbsoluteTimeSeconds"],
    )


def get_acquisition_time(
    con_handle: ConnectionHandler,
) -> tuple[str, float | None]:
    """Interface to get current acquisition time relative to the start
    of acquistion.

    Args:
        con_handle: A unique identifier per mainframe connection.

    Returns:
       Tuple with status and current acquisition time relative to
       the start of acquistion.
    """

    response_json = con_handle.send_request_wait_response(
        "GetAcquisitionTime", None
    )

    if ("AcquisitionTime" not in response_json) or (
        response_json[RETURN_KEY] != GHSReturnValue["OK"]
    ):
        return to_string(response_json[RETURN_KEY], GHSReturnValue), None
    return (
        to_string(response_json[RETURN_KEY], GHSReturnValue),
        response_json["AcquisitionTime"],
    )



# Copyright (C) 2022 Hottinger Bruel and Kjaer Benelux B.V.
# Schutweg 15a
# 5145 NP Waalwijk
# The Netherlands
# http://www.hbm.com

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""``Request`` ``Response`` utilities for JSON-RPC client.

Implementation of request and response utility function for JSON-RPC
client APIs.
"""

import json
from collections import OrderedDict

from .ghsapi_states import RETURN_KEY, GHSReturnValue


def json_rpc_create_request(
    request_id: int, method_name: str, method_param: dict | None
) -> bytes:
    """``Create`` ``JSON-RPC`` client request"""

    if method_param:
        json_obj = OrderedDict(
            [
                ("jsonrpc", "2.0"),
                ("method", method_name),
                ("params", method_param),
                ("id", request_id),
            ]
        )
    else:
        json_obj = OrderedDict(
            [
                ("jsonrpc", "2.0"),
                ("method", method_name),
                ("id", request_id),
            ]
        )
    request_json = json.dumps(json_obj, separators=(",", ":")) + "\0"
    return request_json.encode("utf-8")


def json_rpc_check_errors(request_id: int, response_dict: dict) -> int:
    """Check for errors in JSON-RPC response"""

    if int(response_dict["id"]) != request_id:
        return GHSReturnValue["NOK"]
    # If retreivedRequestID is null means some critical error was
    # received by the client
    if response_dict["id"] == "null":
        try:
            # if response_dict has key "error"
            # Depending on the code the message sent to application can
            # differ
            error_code = response_dict["error"]["code"]
            if error_code == -32700:
                return GHSReturnValue["InvalidJSONFormat"]
            return GHSReturnValue["UnkownErrorMessage"]
        except KeyError:
            return GHSReturnValue["NOK"]
    # Determine how to handle the message depending on what's in it
    try:
        # if response_dict.has_key("error"):
        error_code = response_dict["error"]["code"]
        if error_code == -32601:
            return GHSReturnValue["MethodNotFound"]
        return GHSReturnValue["UnkownErrorMessage"]
    except KeyError:
        return GHSReturnValue["OK"]


def json_rpc_parse_response(
    request_id: int, response_json: bytes | None
) -> dict:
    """``Parse`` ``JSON-RPC`` response"""

    parsed_json = json.loads(response_json[:-1])
    # Check if retrevied JSON has any error code in it
    return_var = json_rpc_check_errors(request_id, parsed_json)
    if return_var != GHSReturnValue["OK"]:
        return {RETURN_KEY: return_var}
    try:
        # if parsed_json has key "result"
        # A result number in the response MUST be a GHSReturnValue
        # integer, return that
        if isinstance(parsed_json["result"], int):
            return {RETURN_KEY: parsed_json["result"]}
        # A result object in the response MUST have a GHSReturnValue
        # named parameter, return that
        try:
            # if parsed_json["result"] has key "GHSReturnValue"
            return parsed_json["result"]
        # Invalid JSON-RPC response
        except KeyError:
            return {RETURN_KEY: GHSReturnValue["NOK"]}
    # Invalid JSON-RPC response
    except KeyError:
        return {RETURN_KEY: GHSReturnValue["NOK"]}



# Copyright (C) 2022 Hottinger Bruel and Kjaer Benelux B.V.
# Schutweg 15a
# 5145 NP Waalwijk
# The Netherlands
# http://www.hbm.com

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""GEN DAQ API - Python.

The GEN DAQ API can be used to control the HBM GEN Series tethered
mainframes without using Perception.

Package Contents:
    Connections: Connection related API functions.
    Acquisition control: Acquisition control related API functions.
"""

from . import acquisition_api as _acquisition
from . import channel_api as _channel
from . import connection_api as _connection
from . import fieldbus_api as _fieldbus
from . import mainframe_api as _mainframe
from . import manage_mainframe_settings as _manage_mainframe_settings
from . import manage_recordings_api as _manage_recordings
from . import recorder_api as _recorder
from .connection import ConnectionHandler
from .ghsapi_states import (
    RETURN_KEY,
    GHSChannelType,
    GHSDigitalOutMode,
    GHSDirection,
    GHSEnableDisable,
    GHSReturnValue,
    GHSSyncStatus,
    GHSTriggerMode,
    GHSUserMode,
)


CLIENT_API_VERSION = 4


class GHS:
    """GEN DAQ API object.

    Attributes:
        _con_handle: An unique identifier per mainframe connection.
    """

    def __init__(self):
        self._con_handle = ConnectionHandler()

    # Connection related API functions.
    def ghs_connect(self, ip_address: int, port_num: int) -> str:
        """Establishes a connection to the mainframe.

        Args:
            ip_address: IP address needs to be an IPV4 address.
            port_num: TCP port number (currently defined as 8006).

        Returns:
            * GHSReturnValue - Connect return status.
        """

        return _connection.connect(
            self._con_handle, ip_address, port_num, CLIENT_API_VERSION
        )

    def ghs_disconnect(self) -> str:
        """Disconnects from a connected mainframe.

        *Results in error when the handle is not valid and / or
        mainframe is not connected.*

        Returns:
            * GHSReturnValue - Disconnect return status.
        """

        return _connection.disconnect(self._con_handle)

    def ghs_get_client_api_version(self) -> int:
        """Client's API version.

        *This function can be useful when ghs_connect() returns
        GHSReturnValue_APIMismatch.*

        Returns:
            * CLIENT_API_VERSION - Version number of the client.
        """

        return CLIENT_API_VERSION

    def ghs_get_current_access(self) -> str:
        """Current access permission to the mainframe.

        *The current access permissions might change at any time.
        For example, when Perception connects concurrently to the same
        mainframe, write access is dropped for the connected GEN DAQ
        API client. If a concurrently connected Perception disconnects,
        write access is granted to the connected GEN DAQ API client.
        API functions that require write access will return
        WriteAccessBlocked in case this client currently has no write
        access.*

        Returns:
            * GHSAccess - Access permission.
        """

        return _connection.get_current_access(self._con_handle)

    # Acquisition control related API functions.
    def ghs_start_preview(self) -> str:
        """Starts preview mode.

        *The system needs to be idle before calling this function.*

        Returns:
            * GHSReturnValue - Start preview status.
        """

        return _acquisition.start_preview(self._con_handle)

    def ghs_stop_preview(self) -> str:
        """Stops preview mode.

        *The system needs to be in preview mode before calling this
        function.*

        Returns:
            * GHSReturnValue - Stop preview status.
        """

        return _acquisition.stop_preview(self._con_handle)

    def ghs_start_recording(self) -> str:
        """Start recording on local storage.

        *The system needs to be idle before calling this function. Note
        that the connected mainframe will generate a recording name.
        This command can be executed only successfully when the local
        storage is set.*

        Returns:
            * GHSReturnValue - Start recording status.
        """

        return _acquisition.start_recording(self._con_handle)

    def ghs_pause_recording(self) -> str:
        """Pauses a started recording.

        *The system needs to be recording before calling this function.*

        Returns:
            * GHSReturnValue - Pause recording status.
        """

        return _acquisition.pause_recording(self._con_handle)

    def ghs_resume_recording(self) -> str:
        """Resumes a paused recording.

        *The system needs to be paused before calling this function.*

        Returns:
            * GHSReturnValue - Resume recording status.
        """

        return _acquisition.resume_recording(self._con_handle)

    def ghs_stop_recording(self) -> str:
        """Stops a started recording.

        *The system needs to be recording before calling this function.*

        Returns:
            * GHSReturnValue - Stop recording status.
        """

        return _acquisition.stop_recording(self._con_handle)

    def ghs_trigger(self) -> str:
        """Issues a trigger.

        Results in error when the trigger cannot be issued.

        Returns:
            * GHSReturnValue - Trigger status.
        """

        return _acquisition.trigger(self._con_handle)

    def ghs_get_acquisition_state(self) -> tuple[str, str | None]:
        """Returns the Acquisition State of the Mainframe.

        Returns:
            * GHSReturnValue - API return status
            * GHSAcquisitionState - acquisition state of the mainframe.
        """

        return _acquisition.get_acquisition_state(self._con_handle)

    def ghs_get_acquisition_start_time(
        self,
    ) -> tuple[str, int | None, int | None, float | None]:
        """Retrieves the absolute time of the start of acquisition.

        Returns:
            * GHSReturnValue - API return status
            * year - The year
            * day - The day number in the year
            * seconds -The number of seconds since midnight.
        """

        return _acquisition.get_acquisition_start_time(self._con_handle)

    def ghs_get_acquisition_time(self) -> tuple[str, float | None]:
        """Retrieves the current acquisition time relative to the start
        of acquistion.

        Returns:
            * GHSReturnValue - API return status
            * acquisition_time - The acquisition time in seconds.
        """

        return _acquisition.get_acquisition_time(self._con_handle)

    # Mainframe APIs

    def ghs_identify(self, identity_flag: bool) -> str:
        """Enable or disable the identification sound of the connected
        mainframe.

        Args:
            identity_flag: Enable or disable flag.

        Returns:
            * GHSReturnValue - Identify API status.
        """

        return _mainframe.identity(self._con_handle, identity_flag)

    def ghs_get_disk_space(self) -> tuple[str, float | None, float | None]:
        """Get total and available mainframe internal disk space.

        Returns:
            * GHSReturnValue - API return status
            * total - Total mainframe internal disk space in GB
            * available - Available internal disk space in GB
        """

        return _mainframe.get_disk_space(self._con_handle)

    def ghs_get_sync_status(self) -> tuple[str, str | None]:
        """Determine the mainframe sync status.

        Returns:
            * GHSReturnValue - API return status
            * GHSSyncStatus - Sync status
        """

        return _mainframe.get_sync_status(self._con_handle)

    def ghs_get_slot_count(self) -> tuple[str, int | None]:
        """Retrieve the number of slots in the mainframe.

        Returns:
            * GHSReturnValue - API return status
            * slot_count - The number of slots in the mainframe
        """

        return _mainframe.get_slot_count(self._con_handle)

    def ghs_get_user_mode(self) -> tuple[str, str | None]:
        """Retrieve the user mode.

        Returns:
            * GHSReturnValue - API return status
            * GHSUserMode - The user mode
        """

        return _mainframe.get_user_mode(self._con_handle)

    def ghs_set_user_mode(self, user_mode: str | int) -> str:
        """Set the user mode.

        *The system needs to be idle before calling this function.*

        Args:
            user_mode: The desired user mode.

        Returns:
            * GHSReturnValue - API return status
        """

        return _mainframe.set_user_mode(self._con_handle, user_mode)

    def ghs_get_mainframe_info(
        self,
    ) -> tuple[str, str | None, str | None, str | None, str | None]:
        """Determine type, name, serial number and firmware version
        information for the connected mainframe.

        *The mainframeType, mainframeName, serialNumber and
        firmwareVersion parameters are UTF-8 encoded.*

        Returns:
            * GHSReturnValue - API return status
            * mainframe_type - The type of the mainframe
            * mainframe_name - The name of the mainframe
            * serial_number - The serial number of the mainframe
            * firmware_version - The firmware version of the mainframe
        """

        return _mainframe.get_mainframe_info(self._con_handle)

    # Manage recordings APIs

    def ghs_delete_all_recordings(self) -> str:
        """Deletes all recordings from local mainframe storage.

        *Recordings are deleted asynchronously.*

        Returns:
            * GHSReturnValue - API return status
        """

        return _manage_recordings.delete_all_recordings(self._con_handle)

    def ghs_delete_last_recording(self) -> str:
        """Deletes the most recent recording from local mainframe
        storage.

        *Recordings are deleted asynchronously.*

        Returns:
            * GHSReturnValue - API return status
        """

        return _manage_recordings.delete_last_recording(self._con_handle)

    def ghs_get_recording_name(self) -> tuple[str, str | None, int | None]:
        """Retrieve the recording base name and recording index of the
        last recording file.

        *The recording base name parameter is UTF-8 encoded.*

        Returns:
            * GHSReturnValue - API return status
            * recording_base_name - The base name of the recording file
            * recording_index - The index of the recording file
        """

        return _manage_recordings.get_recording_name(self._con_handle)

    def ghs_get_storage_location(self) -> tuple[str, str | None]:
        """Retrieve the storage location.

        Returns:
            * GHSReturnValue - API return status
            * GHSStorageLocation - The storage location
        """

        return _manage_recordings.get_storage_location(self._con_handle)

    def ghs_get_high_low_rate_storage_enabled(
        self, source: str | int, slot_id: str
    ) -> tuple[str, str | None, str | None]:
        """Retrieve storage enabled status of high and low rate data
        for the specified recording data source.

        Args:
            source: The recording data source
            slot_id: The slot containing the recorder

        Returns:
            * GHSReturnValue - API return status
            * GHSEnableDisable - Flag to indicate if high rate data is stored
            * GHSEnableDisable - Flag to indicate if low rate data is stored
        """

        return _manage_recordings.get_high_low_rate_storage_enabled(
            self._con_handle, source, slot_id
        )

    def ghs_set_high_low_rate_storage_enabled(
        self,
        source: str | int,
        slot_id: str,
        high_rate_enabled: str | int,
        low_rate_enabled: str | int,
    ) -> str:
        """Enable/disable storage of high and low rate data for the
        specified recording data source.

        Args:
            source: The recording data source
            slot_id: The slot containing the recorder
            high_rate_enabled: Enable/disable storage of high rate data
            low_rate_enabled: Enable/disable storage of low rate data.

        Returns:
            * GHSReturnValue - API return status
        """

        return _manage_recordings.set_high_low_rate_storage_enabled(
            self._con_handle,
            source,
            slot_id,
            high_rate_enabled,
            low_rate_enabled,
        )

    def ghs_set_recording_name(
        self, recording_name: str, recording_index: int
    ) -> str:
        """Set the recording base name and recording index for the next
        recording file.

        *The system needs to be idle before calling this function.
        The recording base name parameter must be UTF-8 encoded.*

        Args:
            recording_name: The desired base name of the recording file
            recording_index: The desired index of the recording file

        Returns:
            * GHSReturnValue - API return status
        """

        return _manage_recordings.set_recording_name(
            self._con_handle, recording_name, recording_index
        )

    def ghs_set_storage_location(self, storage_location: str | int) -> str:
        """Set the storage location.

        *The system needs to be idle before calling this function.*

        Args:
            storage_location: The desired storage location.

        Returns:
            * GHSReturnValue - API return status
        """

        return _manage_recordings.set_storage_location(
            self._con_handle, storage_location
        )

    # Manage recordings APIs

    def ghs_apply_persisted_settings(self) -> str:
        """A mainframe might contain persisted settings (being applied upon
        boot). This method re-applies these settings. In Perception this
        maps on the 'Configured boot' feature.

        *The system needs to be idle before calling this function.
        This function overwrites any previously set settings and / or
        persisted settings.*

        Returns:
            API status.
        """

        return _manage_mainframe_settings.apply_persisted_settings(
            self._con_handle
        )

    def ghs_persist_current_settings(self) -> str:
        """Persists the current mainframe settings.

        *The persisted mainframe settings are applied upon a mainframe
        boot.*

        Returns:
            * GHSReturnValue - API return status
        """

        return _manage_mainframe_settings.persist_current_settings(
            self._con_handle
        )

    def ghs_get_current_settings(self) -> tuple[str, bytes | None, int | None]:
        """Retrieves the current mainframe settings as a blob.

        *As this blob can be of variable size, it is upon the caller to
        ensure reading the correct amount of memory.
        Memory for this blob is allocated by the API.*

        Returns:
            * GHSReturnValue - API return status
            * blob - Pointer to the settings blob
            * blob_size - Size of the settings blob
        """

        return _manage_mainframe_settings.get_current_settings(
            self._con_handle
        )

    def ghs_set_current_settings(self, blob: bytes, blob_size: int) -> str:
        """Persists the current mainframe settings.

        *The persisted mainframe settings are applied upon a mainframe
        boot.*

        Args:
            blob: Settings blob.
            blob_size: Size of the settings blob.

        Returns:
            * GHSReturnValue - API return status
        """

        return _manage_mainframe_settings.set_current_settings(
            self._con_handle, blob, blob_size
        )

    # Recorder APIs

    def ghs_get_channel_count(self, slot_id: str) -> tuple[str, int | None]:
        """Retrieve the number of channels for a recorder.

        *This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder

        Returns:
            * GHSReturnValue - API return status
            * channel_count - The number of channels for the recorder
        """

        return _recorder.get_channel_count(self._con_handle, slot_id)

    def ghs_get_digital_output(
        self, slot_id: str, digital_output: str | int
    ) -> tuple[str, str | None]:
        """Retrieve the Digital Output Mode for a specified Output ID in a
        recorder.

        *This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            digital_output: The output number desired.

        Returns:
            * GHSReturnValue - API return status
            * GHSDigitalOutMode - The digital output mode for that output
        """

        return _recorder.get_digital_output(
            self._con_handle, slot_id, digital_output
        )

    def ghs_get_recorder_enabled(self, slot_id: str) -> tuple[str, str | None]:
        """Determine if recorder is enabled or disabled.

        *This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder

        Returns:
            * GHSReturnValue - API return status
            * GHSEnableDisable - The recorder enabled status
        """

        return _recorder.get_recorder_enabled(self._con_handle, slot_id)

    def ghs_get_recorder_info(
        self, slot_id: str
    ) -> tuple[str, str | None, str | None, str | None, str | None]:
        """Determine type, name, serial number and firmware version
        information for a recorder.

        *The recorderType, recorderName, serialNumber and firmwareVersion
        parameters are UTF-8 encoded.*

        *This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder

        Returns:
            * GHSReturnValue - API return status
            * recorder_type - The type of the recorder
            * recorder_name - The name of the recorder
            * serial_number - The serial number of the recorder
            * firmware_version - The firmware version of the recorder
        """

        return _recorder.get_recorder_info(self._con_handle, slot_id)

    def ghs_get_sample_rate(self, slot_id: str) -> tuple[str, float | None]:
        """Determine the sample rate for a recorder.

        *This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder

        Returns:
            * GHSReturnValue - API return status
            * sample_rate - in samples per second
        """

        return _recorder.get_sample_rate(self._con_handle, slot_id)

    def ghs_set_digital_output(
        self,
        slot_id: str,
        digital_output: str | int,
        digital_output_mode: str | int,
    ) -> str:
        """Set the Digital Output Mode for a specified Output ID in a
        recorder.

        *This method will only process requests from the connected client
        with the most privileges order (Privileges order: 1- Perception,
        2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            digital_output: The output number desired.
            digital_output_mode: The digital output mode to set

        Returns:
            * GHSReturnValue - API return status
        """

        return _recorder.set_digital_output(
            self._con_handle, slot_id, digital_output, digital_output_mode
        )

    def ghs_set_recorder_enabled(
        self,
        slot_id: str,
        enabled: str | int,
    ) -> str:
        """Enable or disable a recorder.

        *The system needs to be idle before calling this function.*

        *This method will only process requests from the connected client
        with the most privileges order (Privileges order: 1- Perception,
        2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            enabled: Set to GHS_Enable/GHS_Disable to enable/disable

        Returns:
            * GHSReturnValue - API return status
        """

        return _recorder.set_recorder_enabled(
            self._con_handle, slot_id, enabled
        )

    def ghs_set_sample_rate(
        self,
        slot_id: str,
        sample_rate: float,
    ) -> str:
        """Set the sample rate for a recorder.

        *The system needs to be idle before calling this function.
        This function overwrites any previously set sample rate setting for
        the specified recorder.
        If the specified sample rate is not supported by the recorder, the
        sample rate is rounded to the nearest supported sample rate.*

        *This method will only process requests from the connected client
        with the most privileges order (Privileges order: 1- Perception,
        2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            sample_rate: In samples per second.

        Returns:
            * GHSReturnValue - API return status
        """
        return _recorder.set_sample_rate(
            self._con_handle, slot_id, sample_rate
        )

    # Channel APIs

    ## Functions

    def ghs_get_channel_type(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, int | None]:
        """Determine the type of a channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * GHSChannelType - Type of the channel
        """

        return _channel.get_channel_type(
            self._con_handle, slot_id, channel_index
        )

    def ghs_get_channel_name(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, int | None]:
        """Determine the name of a channel.

        *The channel name is UTF-8 encoded*

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * channel_name - The name of the channel
        """

        return _channel.get_channel_name(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_channel_name(
        self, slot_id: str, channel_index: int, channel_name: str
    ) -> str:
        """Set the name for a channel.

        *The channel name is UTF-8 encoded*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            channel_name: The name of the channel

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_channel_name(
            self._con_handle, slot_id, channel_index, channel_name
        )

    def ghs_get_channel_storage_enabled(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, str | None]:
        """Determine if storage is enabled or disabled for a channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * GHSEnableDisable - The storage enabled status for the channel
        """

        return _channel.get_channel_storage_enabled(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_channel_storage_enabled(
        self,
        slot_id: str,
        channel_index: int,
        enabled: str | int,
    ) -> str:
        """Enable or disable storage for a channel.

        *The system needs to be idle before calling this function.*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            enabled: The desired storage enabled status for the channel

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_channel_storage_enabled(
            self._con_handle, slot_id, channel_index, enabled
        )

    def ghs_cmd_zeroing(
        self, slot_id: str, channel_index: int, ezeroing: str | int
    ) -> str:
        """Perform zeroing in a channel.

        *The system needs to be idle before calling this function.*

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            ezeroing: Zero / Unzero the specific channel

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.cmd_zeroing(
            self._con_handle, slot_id, channel_index, ezeroing
        )

    ## Analog Module

    def ghs_get_trigger_settings(
        self, slot_id: str, channel_index: int
    ) -> tuple[
        str, str | None, float | None, float | None, float | None, str | None
    ]:
        """Determine the trigger settings for an analog channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * GHSTriggerMode - Trigger Mode
            * primary_level - The primary trigger level
            * secondary_level - The secondary trigger level
            * hysteresis - The trigger hysteresis
            * direction - The trigger direction
        """

        return _channel.get_trigger_settings(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_trigger_settings(
        self,
        slot_id: str,
        channel_index: int,
        trigger_mode: str | int,
        primary_level: float,
        secondary_level: float,
        hysteresis: float,
        direction: str | int,
    ) -> str:
        """Set the trigger settings for an analog channel.

        *The system needs to be idle before calling this function.*

        *This function overwrites any previously set trigger settings
        for the specified recorder.*

        *If the specified trigger mode or value is not supported by the
        recorder, the trigger mode remains unchanged or the value is
        rounded to the nearest supported value.*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            trigger_mode: Trigger Mode. Default is Basic
            primary_level: The desired primary trigger level
            secondary_level: The desired secondary trigger level
            hysteresis: The desired trigger hysteresis
            direction: The desired trigger direction

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_trigger_settings(
            self._con_handle,
            slot_id,
            channel_index,
            trigger_mode,
            primary_level,
            secondary_level,
            hysteresis,
            direction,
        )

    def ghs_get_signal_coupling(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, str | None]:
        """Determine the signal coupling for an analog channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * GHSSignalCoupling - The signal coupling
        """

        return _channel.get_signal_coupling(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_signal_coupling(
        self,
        slot_id: str,
        channel_index: int,
        signal_coupling: str | int,
    ) -> str:
        """Set the signal coupling for an analog channel.

        *If the specified signal coupling mode is not supported by the
        recorder, the signal coupling mode remains unchanged.*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            signal_coupling: The desired signal coupling.

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_signal_coupling(
            self._con_handle,
            slot_id,
            channel_index,
            signal_coupling,
        )

    def ghs_get_input_coupling(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, str | None]:
        """Determine the input coupling for an analog channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * GHSInputCoupling - The input coupling
        """

        return _channel.get_input_coupling(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_input_coupling(
        self,
        slot_id: str,
        channel_index: int,
        input_coupling: str | int,
    ) -> str:
        """Set the input coupling for an analog channel.

        *If the specified input coupling mode is not supported by the
        recorder, the input coupling mode remains unchanged.*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            input_coupling: The desired input coupling.

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_input_coupling(
            self._con_handle,
            slot_id,
            channel_index,
            input_coupling,
        )

    def ghs_get_span_and_offset(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, float | None, float | None]:
        """Determine the span and offset for an analog channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * span - The span in user units
            * offset - The offset in user units
        """

        return _channel.get_span_and_offset(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_span_and_offset(
        self,
        slot_id: str,
        channel_index: int,
        span: float,
        offset: float,
    ) -> str:
        """Set Span and offset for analog channels.

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            span: The span in user units. The value is adapted to available options.
            offset: The offset in user units. The value is adapted to available options.

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_span_and_offset(
            self._con_handle,
            slot_id,
            channel_index,
            span,
            offset,
        )

    def ghs_get_filter_type_and_frequency(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, str | None, float | None]:
        """Determine the filter type and frequency for an analog channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * GHSFilterType - The filter type
            * frequency - The filter frequency in Hz
        """

        return _channel.get_filter_type_and_frequency(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_filter_type_and_frequency(
        self,
        slot_id: str,
        channel_index: int,
        filter_type: str | int,
        frequency: float,
    ) -> str:
        """Set the filter type and frequency for an analog channel.

        *This function overwrites any previously set filter settings
        for the specified recorder.*

        *If a specified filter type or value is not supported by the
        recorder, the filter type remains the same or the value is rounded to
        a supported vale.*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            filter_type: The filter type. Default is Bessel.
            frequency: The filter frequency in Hz.

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_filter_type_and_frequency(
            self._con_handle,
            slot_id,
            channel_index,
            filter_type,
            frequency,
        )

    def ghs_get_excitation(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, str | None, float | None]:
        """Determine the excitation type and value for an analog channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * GHSExcitationType - The excitation type
            * excitation_value - The excitation value in user units (voltage or current).
        """

        return _channel.get_excitation(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_excitation(
        self,
        slot_id: str,
        channel_index: int,
        excitation_type: str | int,
        excitation_value: float,
    ) -> str:
        """Set the excitation type and value for an analog channel.

        *The system needs to be idle before calling this function.*

        *If the specified excitation type or value is not supported by the
        recorder, the excitation type remains unchanged or the value is
        rounded to the nearest supported value.*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            excitation_type: The desired excitation type. Default is Voltage.
            excitation_value: The desired excitation value in user units (voltage or current).

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_excitation(
            self._con_handle,
            slot_id,
            channel_index,
            excitation_type,
            excitation_value,
        )

    def ghs_get_amplifier_mode(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, str | None]:
        """Determine the amplifier mode for an analog channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * GHSAmplifierMode - The amplifier mode
        """

        return _channel.get_amplifier_mode(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_amplifier_mode(
        self,
        slot_id: str,
        channel_index: int,
        amplifier_mode: str | int,
    ) -> str:
        """Set the amplifier mode for an analog channel.

        *The system needs to be idle before calling this function.*

        *If the specified amplifier mode is not supported by the recorder, the
        amplifier mode remains unchanged.*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            amplifier_mode: The desired amplifier mode

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_amplifier_mode(
            self._con_handle,
            slot_id,
            channel_index,
            amplifier_mode,
        )

    def ghs_get_technical_units(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, str | None, float | None, float | None]:
        """Determine the technical units, unit multiplier and unit offset for
        an analog channel.

        *The units parameter is UTF-8 encoded.*

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * units - The technical units (e.g. 'V' for Volt or 'Hz' for Hertz).
            * multiplier - The technical units multiplier value.
            * offset - The technical units offset value.
        """

        return _channel.get_technical_units(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_technical_units(
        self,
        slot_id: str,
        channel_index: int,
        units: str,
        multiplier: float,
        offset: float,
    ) -> str:
        """Set the technical units, unit multiplier and unit offset for an
        analog channel.

        *The units parameter must be UTF-8 encoded.*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            units: The desired technical units (e.g. 'V' for Volt or 'Hz' for Hertz).
            multiplier: The desired technical units multiplier value.
            offset: The desired technical units offset value.

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_technical_units(
            self._con_handle,
            slot_id,
            channel_index,
            units,
            multiplier,
            offset,
        )

    def ghs_get_auto_range(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, str | None, float | None]:
        """Determine the auto range enable and time settings.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * GHSEnableDisable - The auto range enabled setting.
            * auto_range_time - The time for auto range in seconds.
        """

        return _channel.get_auto_range(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_auto_range(
        self,
        slot_id: str,
        channel_index: int,
        auto_range_enabled: str | int,
        auto_range_time: float,
    ) -> str:
        """Set Auto range settings for analog channels.

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            auto_range_enabled: The auto range enabled setting. The value is adapted to available options.
            auto_range_time: The time for auto range in seconds. The value is adapted to available options.

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_auto_range(
            self._con_handle,
            slot_id,
            channel_index,
            auto_range_enabled,
            auto_range_time,
        )

    def ghs_cmd_auto_range_now(
        self,
        slot_id: str,
        channel_index: int,
        auto_range_time: float,
    ) -> str:
        """Command a single shot for auto range.

        *The system needs to be acquiring for this function to have any
        effect.*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            auto_range_time: The time for auto range in seconds. The value is adapted to available options.

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.cmd_auto_range_now(
            self._con_handle,
            slot_id,
            channel_index,
            auto_range_time,
        )

    def ghs_get_channel_cal_info(
        self, slot_id: str, channel_index: int
    ) -> tuple[
        str,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
    ]:
        """Retrieve calibration information for an analog channel.

        *The calibrationDateTime, verificationDateTime,
        powerVerificationDateTime, calibrationLab, verificationLab and
        powerVerificationLab parameters are UTF-8 encoded.*

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * calibration_date_time - The date and time this analog channel has been calibrated.
            * verification_date_time - The date and time the calibration for this analog channel has been verified.
            * power_verification_date_time - The date and time the power calibration for this analog channel has been verified (if applicable).
            * calibration_lab - The laboratory that conducted the calibration for this analog channel.
            * verification_lab - The laboratory that verified the calibration for this analog channel.
            * power_verification_lab - The laboratory that verified the power calibration for this analog channel (if applicable).
        """

        return _channel.get_channel_cal_info(
            self._con_handle, slot_id, channel_index
        )

    ## Timer/Counter Module

    def ghs_get_timer_counter_gate_time(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, float | None]:
        """Determine the gate time for a timer/counter channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * gate_time - The gate time in seconds
        """

        return _channel.get_timer_counter_gate_time(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_timer_counter_gate_time(
        self,
        slot_id: str,
        channel_index: int,
        gate_time: float,
    ) -> str:
        """Determine the gate time for a timer/counter channel.

        The system needs to be idle before calling this function.

        If the specified timer/counter gate time is not supported by the
        recorder, the timer/counter gate time is rounded to the nearest
        supported gate time.

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            gate_time: The desired gate time in seconds.

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_timer_counter_gate_time(
            self._con_handle,
            slot_id,
            channel_index,
            gate_time,
        )

    def ghs_get_timer_counter_mode(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, str | None]:
        """Determine the mode for a timer/counter channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * GHSTimerCounterMode - The timer/counter mode
        """

        return _channel.get_timer_counter_mode(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_timer_counter_mode(
        self,
        slot_id: str,
        channel_index: int,
        mode: str | int,
    ) -> str:
        """Set the mode for a timer/counter channel.

        *The system needs to be idle before calling this function.*

        *If the specified timer/counter mode is not supported by the recorder, the
        timer/counter mode remains unchanged.*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            mode: The desired timer/counter mode. Default is RPMUniDirectional.

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_timer_counter_mode(
            self._con_handle,
            slot_id,
            channel_index,
            mode,
        )

    def ghs_get_timer_counter_range(
        self, slot_id: str, channel_index: int
    ) -> tuple[str, float | None, float | None]:
        """Determine the range for a timer/counter channel.

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel

        Returns:
            * GHSReturnValue - API return values
            * lower_value - The lower range value.
            * upper_value - The upper range value.
        """

        return _channel.get_timer_counter_range(
            self._con_handle, slot_id, channel_index
        )

    def ghs_set_timer_counter_range(
        self,
        slot_id: str,
        channel_index: int,
        lower_value: float,
        upper_value: float,
    ) -> str:
        """Set the range for a timer/counter channel.

        *The system needs to be idle before calling this function.*

        *If the specified timer/counter range is illegal (i.e. upperValue <
        lowerValue), the timer/counter range values are corrected to the
        nearest possible values.*

        *ReadWrite - This method will only process requests from the
        connected client with the most privileges order (Privileges
        order: 1- Perception, 2- GenDaq, 3- Other)*

        Args:
            slot_id: The slot containing the recorder
            channel_index: The zero-based index of the channel
            lower_value: The desired lower range value.
            upper_value: The desired upper range value.

        Returns:
            * GHSReturnValue - API return values
        """

        return _channel.set_timer_counter_range(
            self._con_handle,
            slot_id,
            channel_index,
            lower_value,
            upper_value,
        )

    ## Fieldbus

    def ghs_get_fieldbus_data_name_and_unit(
        self,
        data_index: int,
    ) -> tuple[str, str | None, str | None]:
        """Gets the name of specific field bus data value, by giving the index of the data value.

        *The number of data values can be retrieved by ghs_get_fieldbus_data_count().*

        *In case of no formulas (or reserved values acquisition state and latency) published,
        an error code is returned: FieldBusError_NoFormulasDeployed.*

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            data_index: Zero-based index of the data value.

        Returns:
            * GHSReturnValue - API return values
            * data_name - Name of the data value.
            * data_unit - Unit of the data value.
        """

        return _fieldbus.get_fieldbus_data_name_and_unit(
            self._con_handle,
            data_index,
        )

    def ghs_get_fieldbus_data_count(
        self,
    ) -> tuple[str, int | None]:
        """Gets the number of data that are configured to be sent through the field bus, excluding the timestamp.

        *The number of data values can be retrieved by ghs_get_fieldbus_data_count().*

        *Read - This method can be called by multiple connected clients at same
        time.*

        Returns:
            * GHSReturnValue - API return values
            * data_count - The number of data that are selected to be published, excluding timestamp.
        """

        return _fieldbus.get_fieldbus_data_count(
            self._con_handle,
        )

    def ghs_get_fieldbus_request_snapshot(
        self,
        data_count: int,
    ) -> tuple[str, list[float] | None, int | None, list[float] | None]:
        """Request a single snapshot of the field bus data.

        *In this case the timestamp received is not grid aligned.*

        *If the system is not acquiring the method returns an error code: SystemNotRecording*

        *In case of no formulas you get the error return code back: FieldBusError_NotConfigured*

        *If the field bus is already enabled then this method is not allowed. 
        An error return code is returned: FieldBusError_FieldBusAlready_Enabled*

        *If number of elements allocated by the customer are less than the actual formulas number, 
        then we return an error code.*

        *Read - This method can be called by multiple connected clients at same
        time.*

        Args:
            data_count: The number of data that are selected to be published, excluding timestamp. 
        Returns:
            * GHSReturnValue - API return values
            * timestamp -  Timestamp of the FieldBus data relative to start of acquisition (not grid aligned)
            * data_count - The number of data that are selected to be published, excluding timestamp.
            * data - Array of floats that contains the data
        """

        return _fieldbus.get_fieldbus_request_snapshot(
            self._con_handle,
            data_count
        )



# Copyright (C) 2022 Hottinger Bruel and Kjaer Benelux B.V.
# Schutweg 15a
# 5145 NP Waalwijk
# The Netherlands
# http://www.hbm.com

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implementaion of Connection module."""

import errno
import socket
from struct import pack, unpack

from . import json_rpc
from .ghsapi_states import RETURN_KEY, GHSReturnValue

MAX_CONNECTIONS = 30


class ConnectionHandler:
    """A unique identifier per mainframe connection.

    It is used by all API calls to distinguish between mainframes.

    Attributes:
        connection_count: An integer count of all connections.
        api_version_header: Client API header version.
        request_id: Client request id
        sock: Socket object
        ip_address: Mainframe ip address
    """

    connection_count = 0
    api_version_header = 1195638785

    def __init__(self):
        self.request_id = 0
        self.sock = 0
        self.ip_address = 0

    def get_num_of_connections(self) -> int:
        """Get count of all connections."""

        return self.connection_count

    def get_ip_address(self) -> int:
        """Get ip address of mainframe."""

        return self.ip_address

    def connection_establish(self, ip_address: int, port_num: int) -> int:
        """Establishes connection to the mainframe.

        Args:
            ip_address: IP address of the mainframe.
            port_num: Mainframe port number.

        Returns:
            Integer value representing connection status code.
        """

        if not ip_address or not port_num:
            return GHSReturnValue["NullPtrArgument"]
        if self.get_num_of_connections() > MAX_CONNECTIONS:
            return GHSReturnValue["ConnectionFailed"]

        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except socket.error:
            return GHSReturnValue["ConnectionFailed"]
        except RuntimeError:
            return GHSReturnValue["NOK"]

        try:
            self.sock.connect((ip_address, port_num))
        except socket.gaierror:
            return GHSReturnValue["ConnectionFailed"]
        except socket.error:
            return GHSReturnValue["NoConnection"]
        except RuntimeError:
            return GHSReturnValue["NOK"]

        self.ip_address = ip_address
        self.connection_count += 1
        return GHSReturnValue["OK"]

    def send_request_wait_response(
        self, method_name: str, method_param: dict | None
    ) -> dict:
        """Sends request to the mainframe.

        Args:
            method_name: Request method name .
            method_param: Request method parameter.

        Returns:
            Dict representing response from the mainframe.
        """

        if not method_name:
            return {RETURN_KEY: GHSReturnValue["NullPtrArgument"]}

        self.request_id += 1
        request_json = json_rpc.json_rpc_create_request(
            self.request_id, method_name, method_param
        )
        write_len = len(request_json)
        header_sx = pack("!I", write_len) + pack("!I", self.api_version_header)

        try:
            if self.connection_write(header_sx, len(header_sx)) != len(
                header_sx
            ):
                return {RETURN_KEY: GHSReturnValue["NOK"]}
            if self.connection_write(request_json, write_len) != write_len:
                return {RETURN_KEY: GHSReturnValue["NOK"]}
        except OSError:
            return {RETURN_KEY: GHSReturnValue["NoConnection"]}
        except RuntimeError:
            return {RETURN_KEY: GHSReturnValue["NOK"]}
        except Exception:
            return {RETURN_KEY: GHSReturnValue["NoConnection"]}

        header_rx_size = 8
        header_rx = self.connection_read(header_rx_size)
        if header_rx_size != len(header_rx):
            return {RETURN_KEY: GHSReturnValue["NOK"]}

        header_rx_packet_header = unpack(">II", header_rx)
        if header_rx_packet_header[1] != self.api_version_header:
            return {RETURN_KEY: GHSReturnValue["NOK"]}

        response_json = self.connection_read(header_rx_packet_header[0])
        return json_rpc.json_rpc_parse_response(self.request_id, response_json)

    def connection_read(self, length: int) -> bytes:
        """Read message in bytes.

        Args:
            length: Message length.

        Returns:
            Bytes of message read.

        Raises:
            OSError: When socket not connected
            RuntimeError: When socket connection broken
        """

        message = b""
        while len(message) < length:
            try:
                packet = self.sock.recv(length - len(message))
                if not packet:
                    return None
                message += packet
            except socket.error as socket_error:
                err = socket_error.args[0]
                if err in (errno.EAGAIN, errno.EWOULDBLOCK):
                    print("connection_read: No data available")
                # return None
                # print("connection_read: No data available")
                # return None
        return message

    def connection_write(self, message: bytes, length: int) -> int:
        """Writes message in bytes.

        Args:
            message: Message to write.
            length: Message length.

        Returns:
            Integer representing bytes written.

        Raises:
            OSError: When socket not connected
            RuntimeError: When socket connection broken
        """

        written_bytes = 0
        sent_bytes = 0

        while written_bytes < length:
            try:
                sent_bytes = self.sock.send(message[written_bytes:])
            except AttributeError as no_socket:
                raise OSError("Socket not Connected") from no_socket
            except socket.gaierror as no_socket:
                raise OSError("Socket not Connected") from no_socket
            except RuntimeError as un_specific_error:
                raise RuntimeError from un_specific_error
            except Exception as any_exception:
                raise Exception from any_exception
            if sent_bytes == 0:
                raise RuntimeError("Socket Connection Broken")
            written_bytes = written_bytes + sent_bytes
            sent_bytes = 0

        return written_bytes
